/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>

#include <string.h>

#include <qdatetime.h>
#include <qfile.h>
#include <qtimer.h>

#include <kuniqueapplication.h>

#include "ar_client.h"
#include "ar_parser.h"
#include "bt_comm.h"
#include "logger.h"

extern logger2*     	    theLog;
extern AR_Parser*   	    conf;
extern KUniqueApplication*  pApp;
extern KAnyRemote*  	    mainWin;

AR_Client::AR_Client()
{
    //theLog->debug("AR_Client::AR_Client");
    callerId 	    = QString("NO CALLER ID");
    callOn          = false;
    model   	    = MODEL_DEFAULT;
    modifierString  = "";
    remoteOn	    = true;
    repeatCmdPtr    = NULL;	
    repeatNow       = false;
    useCallId	    = true;
    serverMode      = SERVER_NO;
    initialized     = 0;
    timers    	    = NULL;
    serverMode	    = 0;
    paramValue	    = -1;
}


AR_Client::~AR_Client()
{
}

bool AR_Client::isConnected()
{
    return Comm.isConnected();
}

bool AR_Client::isServer()
{
    return (serverMode > 0);
}

type_key*  AR_Client::getAutoRepeatCmdIndex()
{
    return repeatCmdPtr;
}


void AR_Client::setStyle(QString device)    	// Determine is we were run as server or as client
{   	    	    	    	    	 
    // device should be in format:
    // 1. /dev/something
    // 2. socket:5000
    // 3. bluetooth:10 or bluetooth:default (port will be 10 by default)
    
    if (device.contains(INET_SOCKET)) {
        port = (device.mid(QString(INET_SOCKET).length())).toInt();
        theLog->debug("TCP Server mode. Will use port " + QString().setNum(port));
    	serverMode = SERVER_TCP;
    } else if (device.contains(BT_SOCKET)) {
        port = 10;
        QString chan("10");
        if (!device.contains(":")) {
            chan = device.mid(QString(BT_SOCKET).length()+1);
            port = chan.toInt();
        }
        theLog->debug("Bluetooth Sockets Server mode. Will use channel " + chan);
    	serverMode = SERVER_BT;
    } else {
    	serverMode = SERVER_NO;
	port = -1;
    }
}

bool AR_Client::initServerPort()    	    // If acts as server then open socket
{   	    	    	    	    	    // if acts as client do nothing
    if (serverMode == SERVER_TCP) {
    	theLog->info("initPort() opening TCP socket");
        if (Comm.openSocketPort(SERVER_TCP, port) < 0) {
            return false;
        }
    } else if (serverMode == SERVER_BT) {
    	theLog->info("initPort() opening BT socket");
    	if (Comm.openSocketPort(SERVER_BT, port) < 0) {
            return false;
        }
    } else {
    	theLog->info("initPort() nothing to do");
    }
    
    return true;
}

void AR_Client::setInitialized() {
    if (initialized == 0) { 
    	// Run "onInit" command
    	int i = 0;
    	type_key *k = conf->findItem(QString(EVT_INIT), i);
    	if (k && i == 1) {
    	    theLog->info("Exec cmd on init");
    	    handleCmdByKey(k);  // Since we not connected yet we did not interested in return code
    	} else {
    	    theLog->info("No command on init\n");
    	}
    	initialized++;
    }
}

int AR_Client::listenForConnection() {
   
    if (serverMode == SERVER_NO) {
    	return -1;
    }   

    int ret = Comm.listenAndAcceptSocketConn(serverMode);
    
    if (ret == 1) { 	    	// Connection established
    	int i = 0;
    	type_key* tk = conf->findItem(EVT_CONNECT, i);
    	if (tk != NULL && i == 1) {
    	    handleCmdByKey(tk);
    	} 
    }
    return ret;
}

// Used only in client style
bool AR_Client::connectToBT(QString port)
{
    //theLog->debug("AR_Client::connectToBT");
    
    if (Comm.initPort(port)) {
    	theLog->info("Modem initializing");
 
    	QString answer;
    	Comm.sendCommand("AT\r", answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString(" AT -> ") + answer);
    	    Disconnect();
    	    return false;
    	}

    	Comm.sendCommand("ATE0\r", answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("ATE0 -> ") + answer);
    	    Disconnect();
    	    return false;
    	}

    	QString chrset = conf->getCharset();
    	if (!(chrset.isNull() || chrset.isEmpty())) {
    	    Comm.sendCommand(QString(DEF_AT_CSCS) + "\"" + chrset + "\"\r",answer, QString::null);
    	    if (answer.contains("OK") == 0) {
    		theLog->error(QString("Can't set charset to ") + chrset);
    	    }		
    	}
    	pApp->processEvents();

    	Comm.sendCommand(QString(DEF_AT_CLIP) + "\r",answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("Can't set CLIP ON"));
    	}
    	pApp->processEvents();

    	QString cmer = conf->getAT_CMER(CMER_ON);
    	if (cmer.isNull()) {		// We have to determine model
    	    theLog->debug("Determine phone model");
    	
    	    Comm.sendCommand(QString(DEF_AT_CGMI) + "\r", answer, QString::null);

    	    if (answer.contains(STR_MOTOROLA)) {
    		model = MODEL_MOTOROLA;
    		cmer  = DEF_AT_CMER_ON_MOTOROLA;
    	    } else if (answer.contains(STR_SE)) {
    		model = MODEL_SE;
    		cmer  = DEF_AT_CMER_ON_SE;
    	    } else if (answer.contains(STR_SAGEM)) {
    		model = MODEL_SAGEM;
    		cmer  = DEF_AT_CMER_ON_SAGEM;
    	    } else if (answer.contains(STR_SIEMENS)) {
    		model = MODEL_SIEMENS;
    		cmer  = DEF_AT_CMER_ON_SIEMENS;
    	    } else {
    		model = MODEL_DEFAULT;
    		cmer  = DEF_AT_CMER_ON_DEFAULT;
    	    }
    	} 
    	cmer += "\r";

    	Comm.sendCommand(cmer, answer, QString::null);
    	if (answer.contains("OK") == 0) {
    	    theLog->error(QString("AT_CMER SET >") + cmer + "< ->" + answer);
    	    Disconnect();
    	    return false;
    	}
    	pApp->processEvents();

    	// Siemens S55 needs additional AT+CMEC=2 to make AT+CKPD works
    	// not sure about other Siemens phones
    	if (model == MODEL_SIEMENS) {
    	    Comm.sendCommand(QString(DEF_AT_CMEC) + "\r", answer, QString::null);
    	    if (answer.contains("OK") == 0) {
    		theLog->error(QString("Failed in AT+CMEC"));
    	    }
    	}

    	// Will set global callerId
    	int ret = Comm.getClip(callerId);
    	if(ret == EXIT_NOK) {		// Got ERROR on AT+CLCC; probably phone does not supports this command
    	    theLog->error(QString("fails in getClip"));
    	    useCallId = false;
    	}
    }
    
    setInitialized();

    manageAlarms(ALARM_CLEAN);

    if (serverMode == SERVER_NO) {
    	// These is not fatal errors
    	if (Comm.sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	    theLog->debug("Fails in sendSeq(getToMainMenu())");
    	}
    }
    pApp->processEvents();

    theLog->info("Connection established");
    //setRemoteOn(true);

    int i = 0;
    type_key* tk = conf->findItem(EVT_CONNECT, i);
    if (tk != NULL && i == 1) {

    	int ret = handleCmdByKey(tk);
    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    return true;    // Return true because connection already established
    	}
    } else {
    	theLog->info("Start event forwarding");
    }

    return true;
}

// This is not clean-up function. It just finished the connection.
void AR_Client::Disconnect()
{
    theLog->info("Finish events forwarding");

    pApp->processEvents();

    if (serverMode == 0) {
        QString answer;
        QString cmer = conf->getAT_CMER(CMER_OFF);
        if (cmer.isNull()) {
   	    if (model == MODEL_MOTOROLA) {
	        cmer = DEF_AT_CMER_ON_MOTOROLA;
    	    } else if (model == MODEL_SE) {
	        cmer = DEF_AT_CMER_OFF_SE;
    	    } else if (model == MODEL_SAGEM) {
	        cmer = DEF_AT_CMER_OFF_SAGEM;
    	    } else if (model == MODEL_SIEMENS) {
	        cmer = DEF_AT_CMER_OFF_SIEMENS;
	    } else {
	        cmer = DEF_AT_CMER_ON_DEFAULT;
	    }
    
        }
        cmer += "\r";
    
        Comm.sendCommand(cmer, answer, QString::null);
        //handleCommand((char *)answer.ascii()); // Does it needed ?
    
        // and resend No back
        //Comm.sendCommand("AT+CKPD=\"e\"\r", answer);  seems it is not obligatory
    } 
    
    pApp->processEvents();
    Comm.closePort(1);
    
    return;
}

QString AR_Client::readResultsFile() 
{
    FILE *fp;
    char tmptext[2048];
    
    fp=fopen(TOFILE,"r");
    if (fp == NULL) return NULL;
    
    int n = fread(tmptext, sizeof(char), 2047, fp);
    tmptext[n] = '\0';
    
    return QString(tmptext);
}

int AR_Client::processCommands(cmdItem* cmdList) 
{
    while (cmdList) {
            
        int ret = execCmd(cmdList->type, cmdList->descr);
        if (ret != EXIT_OK) {
            theLog->debug("Return code from execCmd() is " + QString().setNum(ret));
 	     
            if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
  	        theLog->debug("Clean auto repeat flag after execCmd()");
 	        repeatNow    = 0;
 	        repeatCmdPtr = NULL;
		
	        return ret;
            }
        }
        cmdList = cmdList->next;
    }

    return EXIT_OK;
}

void AR_Client::handleDisconnect(bool needClose)
{
    theLog->debug("handleDisconnect");

    int i;
    type_key *tk = conf->findItem(EVT_DISCONNECT, i);
    if (tk != NULL) {
    	i = handleCmdByKey(tk);
	// No additional handling since we already disconnected
    }
    
    if (needClose) {	    // in server style we wan't do this
    	Comm.closePort(1);
    }
    return;
}

int AR_Client::handleCmdByKey(type_key *tk)
{
    if(tk == NULL) {
    	theLog->debug("handleCmdByKey() NULL input");
    	return EXIT_OK;
    }
    theLog->debug("handleCmdByKey() >" + tk->key + "<");
    
    cmdItem* cmdList = conf->getCommand(tk);
    if (cmdList == NULL) {
        theLog->debug("No appropriate command was found by getCommand()");
    } else {
     	theLog->debug("Ready to execute command");
     
        // In AT mode: User had pressed some key ... 
        // Send ToMainMenu sequence of CKPD's to show main menu again
        if (serverMode == SERVER_NO) {
            if (Comm.sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	        theLog->debug("Fails in sendSeq()");
            }
        }
        
        return processCommands(cmdList);
    }	
	
    return EXIT_OK;	
}

int AR_Client::handleKeyPress(char* keyIn) 
{
    int ret;
    
    theLog->debug(QString("AR_Client::handle_key_press() >") + keyIn + "<");
    
    // Search an alias
    QString key;
    if((key = conf->findAlias(keyIn)).isNull()) {
        key = keyIn;
    }

    // If modifier was set already it needs to verify which multi-key command we collect
    if (!modifierString.isEmpty()) {
    	modifierString += " "; // Add one space first
    	modifierString += key;
    	theLog->debug("Modifier+sequence is: >" + modifierString + "<");
    }
    
    // Execute appropriate command
    int i = 1;
    type_key *tk = conf->findItem(modifierString.isEmpty() ? QString(key) : modifierString, i); 
    if (tk != NULL && i == 1) {
    	theLog->debug("One-key command found");
	
    	if ((!repeatNow) && conf->getAutoRepeat()) {
 	    theLog->debug("Set auto repeat flag");
 	    repeatNow    = true;
 	    repeatCmdPtr = tk;
     	} 
	
	ret = handleCmdByKey(tk);
	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    return ret;
    	}
	
	// AR_Client is not successor of QObject so send timeout to mainWin
	if (conf->getAutoRepeat()) {
	    theLog->debug("Set auto repeat timer");
	    QTimer::singleShot(1000, mainWin, SLOT(autoRepeat()));
	}
	    
    	// Clean modifier string
    	modifierString = "";

    } else if (tk != NULL && i == 0) {  // A part of multi-key command
        theLog->debug("Part of multi key command found");
	
    	if (modifierString.isEmpty()) {
	    theLog->debug("Start of multi key sequence");
    	    modifierString = key;
    	} 
	    
        if(serverMode == SERVER_NO) {
            ret = Comm.sendSeq(conf->getToMainMenu());
	    if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	        return ret;
    	    }
        }

    } else {
	// User had pressed some key ... 
    	// Send ToMainMenu sequence of CKPD's to show main menu again
    	if (serverMode == SERVER_NO && Comm.sendSeq(conf->getToMainMenu()) != EXIT_OK) {
    	    theLog->debug("Fails in sendSeq()");
    	}

        theLog->debug("No approprite key definition was found by findItem()");
	
    	if (!modifierString.isEmpty()) {
	    // Clean modifier string
    	    modifierString = "";
	    theLog->debug("Clean modifier string");
    	}
    }

    return EXIT_OK;
}

// Should be called to handle inpit from BT device
int AR_Client::handleCommand(char* cmd) 
{
    char *prev, *next, *key;
    int ret, i;

    paramValue = -1;
    
    //skip lines starting with \n and \r
    if (cmd[0] == '\r') cmd++;
    if (cmd[0] == '\n') cmd++;

    theLog->debug(QString("AR_Client::handle_command() >") + cmd + "<");
  
    // most common case
    if (!cmd[0]) {
    	return EXIT_OK;
    }
    
    // if received multiline command - handle line by line and return
    prev = cmd;
    next = strchr(cmd, '\r');
    if (next == NULL && serverMode > 0) {   // Java client wiil send +CKEV: 1,1; +CKEV: 1,0
    	next = strchr(cmd, '\r');
    }
    if (next) {
    	theLog->debug("handle_command() multiline");
	
        char copy[1024];
        int len;
    
        do {     
            len = next-prev;
            
            if (len > 2) {
                memcpy(copy, prev, len);
                copy[len] = 0;
                
                // use recursion
                ret = handleCommand(copy);
	    	if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
	    	    return ret;
	    	}
            }
            prev = next+1;
        } while ((next = strchr(prev, '\r'))) ;
    
        // now return
        return EXIT_OK;
    }
    if (strstr(cmd, "OK") != NULL) {	// OK - nothing to do
        return EXIT_OK;
    }
    
    // Handle different keypresses
    // Motorola: +CKEV: "1",1
    // SE      : +CKEV: 1,1
    // Sagem: in rare ) cases it could be (between > and <) >,1:< ?
    if (memcmp(cmd, "+CKEV:", 6) == 0) {
    	theLog->debug("Got +CKEV");
	
	int isSE = MODEL_SE;
	
        key = cmd+7;

	// Sagem test
	//key+=3;

	if (*key == '"') {	// Seems this is Motorola
	    key++;
	    isSE = MODEL_MOTOROLA;
	}

	if(*key == ',') {	// Seems this is Sagem
	    theLog->debug("+CKEV is in rare Sagem format (Empty code)");
	    isSE = MODEL_SAGEM;
	}

	i = 1;
	if (isSE == MODEL_SE) {	// SonyEricsson
	    while(key[i] != ',') i++;
	    key[i] ='\0';
	    
	    i = i+1;
	} else if (isSE == MODEL_MOTOROLA) {	// Motorola
	    while(key[i] != '"') i++;
	    key[i] ='\0';

	    i = i+2;
	} else if (isSE == MODEL_SAGEM) {	// rare Sagem case
	    *key ='\0';
     	} else {
	    theLog->error("Can't recognize +CKEV event !!!");
	    return EXIT_NOK;
	}
    	
	if (key[i] == '1') {
	    theLog->debug("Button down event");
            return handleKeyPress(key);
    	} else if (key[i] == '0') {
	    theLog->debug("Button up event");
            // In general we skip this
	    if (repeatNow) {
	    	// Clear auto repeat flag
		theLog->debug("Clean auto repeat flag");
    	    	repeatNow    = false;
    	    	repeatCmdPtr = NULL;
	    }
	    return EXIT_OK;
    	} else {
    	    theLog->debug(QString("something wrong: key should be pressed or released: cmd=>") + key + "<");
            return EXIT_NOK;
    	}
    } else if (memcmp(cmd, DEF_CLCC,6) == 0) {
	theLog->debug("Caller ID received. Skip it.");
	return EXIT_OK;
    } else if (strstr(cmd, DEF_RING) != NULL) {     // Incoming call
	theLog->info("Incoming call");     	    // This event sent periodically until used answered a call
	
	if (!callOn) {
	    if (useCallId) {
	    	theLog->debug("Get CallerId on incoming call");
    	    	ret = Comm.getClip(callerId);

		theLog->debug("CallerId is " + callerId);
		
	    	if (ret == EXIT_EXACT) { // Active call exists
		    theLog->debug("Caller ID received");
	    	} else { 
		    theLog->debug("Can't get Caller ID. Probably call was finished");
		}

	    }
	    //
	    // May be we can get +CLIP: "+xxxxxxxxxxxx",145,,,,0< here ?
	    // 
            
            i = 0;
	    type_key *tk = conf->findItem(EVT_INCALL, i);
	
    	    if (tk != NULL && i == 1) {
	    	theLog->info("Exec cmd on incoming call");
  	    	ret = handleCmdByKey(tk);
	    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return ret;
    	    	}
	    }
	
	    	
	    // If there is (EndCall) command,then set timer to determine when call ends
	    i = 0;
            tk = conf->findItem(EVT_ENDCALL, i);
	    if (tk != NULL && i == 1) {
	    	theLog->debug("Set call timer");
	    	callOn    = true;
		    
    	    	// AR_Client is not successor of QObject sosend timeout to mainWin
    	    	theLog->debug("Set CallerID timer");
    	    	QTimer::singleShot(1000, mainWin, SLOT(getCallerID()));
	    }
	    
	    return EXIT_OK;
    	} else {
	    theLog->debug("Skip it.");      // This event sent periodically until used answered a call
    	}
    } else if (isBemusedCommand(cmd)) {    // Bemused server emulation tricks
    	return EXIT_OK;
    }
    
    theLog->debug(QString("Unhandled cmd ") + cmd);
    return EXIT_OK;
}

int AR_Client::readCommand(QString& buffer, int l) 
{
    int ch, len = 0;
    //theLog->debug("AR_Client::readCommand()");
    
    if (serverMode) {
        len = Comm.readSocketConn(buffer);
        if (len == EOF) {
    	    theLog->debug("AR_Client::readCommand() EOF");
            return EOF;
	}
    	//theLog->debug(QString("AR_Client::readCommand() >") + buffer + "<");
    } else {
        while (len < l) {
            ch = Comm.readchar(100);
            if (ch >= 0 && ch != '\r') {
	    	buffer += ch;
                len++;
            } else  if (ch == EOF) {
    	        buffer = QString::null;
                return EOF;
            } else {
                break;
            }
        }
    }

    return len;
}

// mode: 0 - clear alarms
//       1 - if fired - process, then clear alarms
void AR_Client::manageAlarms(int mode)
{
    //theLog->debug("AR_Client::manageAlarms");

    QString pf;
    cmdItem *pg;
    int i;
    
    int n = conf->getNumAlarms();
    for(i=0;i<n;i++) {
    	pf = conf->getAlarmInfo(i, &pg);
	
    	// If file exists
	QFile fs(pf);
	if (fs.exists()) {
	    if (mode == ALARM_CLEAN) {
		theLog->debug("Remove file");
	    } else {    	
    	    	theLog->info(QString("Alarm fired (") + pf + ")");
                
	        processCommands(pg);
	    }
	    fs.remove();
	}
    }
    return;
}

int AR_Client::execInFork(const char *cmd)
{
    int ret;
    int cpid = fork();
    if( cpid < 0) {
    	theLog->error("Can not fork!");
    	return -1;
    }

    if(cpid) {	    // father
    	waitpid(cpid,&ret,0);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
    	}
    } else {	    // child
    	theLog->info("Close port in child");
    	Comm.closePort(0);  	
    	ret = system(cmd);
    	if(ret!=-1 && WIFEXITED(ret)) {
	    ret = WEXITSTATUS(ret);
	}
    	exit(ret);
    }

    return ret;
}

int AR_Client::execSimpleCmd(QString command, int toFile)
{
    if (command.isNull() || command.isEmpty()) { // Empty command. Skip it.
        theLog->info("execSimpleCmd null input");
    	return EXIT_OK;
    }
    theLog->info("execSimpleCmd >" + command + "<");
    
 
    if (remoteOn) {
    	int papameterized = command.find(PARAM_CALLID);
    	QString cmd;
	
	if (papameterized >= 0) {   	    	// Need to substitute it with callerID
	    theLog->debug("parametrized command - callerID");
	    // First part
	    cmd = command.left(papameterized) + " ";
	   
	    // Add params and second part
	    cmd = cmd + callerId + command.right(command.length() - papameterized - QString(PARAM_CALLID).length());
	} else if ((papameterized = command.find(PARAM_TIME)) >= 0) {   	    	// Need to substitute it with time
	    theLog->debug("parametrized command - Time");
	    // First part
	    cmd = command.left(papameterized) + " ";
	    	
	    QString time = QTime::currentTime().toString().left(5);
		
	    // Add params and second part
	    cmd = cmd + time + command.right(command.length() - papameterized - QString(PARAM_TIME).length());
    	} else if ((papameterized = command.find(PARAM_MODE)) >= 0) {   	    	// Need to substitute it with mode
    	    theLog->debug("parametrized command - Mode");
	    
	    // First part
    	    cmd = command.left(papameterized);
	    
    	    // Add params and second part
    	    cmd = cmd + conf->getMode() + command.right(command.length() - papameterized - QString(PARAM_MODE).length());	

    	} else if ((papameterized = command.find(PARAM_PARAM)) >= 0) {
            // This is dirty trick to make Bemused commands like VOLM<volume> works
            
            if (paramValue == -1) {
                theLog->error("Parameter value was not set!");
                return EXIT_OK;
            }
	    
	    // First part
	    cmd = command.left(papameterized);
	    
 	    // Add params
	    cmd = cmd + QString().setNum(paramValue);
	    
            paramValue = -1; // Reset value after use
	    //papameterized += strlen(PARAM_PARAM);
	    
    	    // Add second part
    	    cmd = cmd + command.right(command.length() - papameterized - QString(PARAM_PARAM).length());	
            
    	} else {    // Just copy
	    
    	    theLog->debug("non-parametrized command");
    	    cmd = command;
	}
	
	if (toFile) {       // We needs to wait results, so do not run in background
	    cmd += QString(" > ") + TOFILE;
	} else {
	    if (!cmd.contains("&")) {
	        cmd += QString(" &");
	    }
	}
    	theLog->debug("final cmd >" + cmd + "<");
	
    	return (execInFork(cmd.ascii()) == -1 ? EXIT_NOK : EXIT_OK);
    } else {
    	theLog->debug("Remote if OFF. Skip event");
    }
    
    return EXIT_OK;
}

int AR_Client::processSendCmd(QString origCmd)
{
    int ret = EXIT_OK;
    
    QString tag  = origCmd.section(",",0,0).simplifyWhiteSpace();
    QString cmd  = origCmd.section(",",1).simplifyWhiteSpace();
    
    if (tag.isNull() || tag.isEmpty() || cmd.isNull() || cmd.isEmpty()) {
        theLog->debug("processSendCmd(): incorrect input (1)");                
        return EXIT_NOK;
    }
    
    theLog->debug("processSendCmd() is >" + tag + "< >" + cmd + "<");
    
    if (tag == SEND_STRING) {
        ret = Comm.writeSocketConn(cmd);
        if (ret != EXIT_OK) {
	    theLog->debug("Fails in writeSocketConn()");
        }
    } else if (tag == SEND_BYTES) {
        ret = Comm.writeBytesSocketConn(cmd);
        if (ret != EXIT_OK) {
	    theLog->debug("Fails in writeBytesSocketConn()");
        }    
    } else {
        theLog->debug("processSendCmd(): incorrect input (3)");                
        return EXIT_NOK;
    }
    return ret;
}

int AR_Client::execCmdAndGetResults(int type, QString commandOrig, timerCmd* timer) 
{
    theLog->debug("execCmdAndGetResults");

    if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != QString::null) {
       
        QString tag  = commandOrig.section(",",0,0).simplifyWhiteSpace();
        QString cmd  = commandOrig.section(",",1,1).simplifyWhiteSpace();
        
        if (tag.isNull() || tag.isEmpty() || cmd.isNull() || cmd.isEmpty()) {
            theLog->debug("execCmdAndGetResults(): incorrect input");                
            return EXIT_NOK;
        }
        if (type != ID_SET && type != ID_SEND ) {
            theLog->debug("execCmdAndGetResults(): incorrect input (2)");                
            return EXIT_NOK;
        }
        
        int ret = execSimpleCmd(cmd,1);
        if (ret != EXIT_OK) {
 	    theLog->debug("Return code from execSimpleCmd() is " + QString().setNum(ret));
	    return ret;
        }    
        
        QString res = readResultsFile();
	theLog->debug("readResultsFile() " + res);
	QString res2;
        int braces = 1;

        if (type == ID_SET) {
            res2 = CMD_SET;
        } else {            // ID_SEND
            braces = 0;
        }
	
        if (braces) {
            res2 += "(" + tag + "," + res + ");";
	} else {
            res2 = res;
        }
        int needToWrite = 1;
        if (timer != NULL && timer->lastValue != QString::null && timer->lastValue == res2) {
            theLog->debug("No needs to send the same data.");
            needToWrite = 0;            // no needs to send exactly the same string
        }
        if (needToWrite) {
            if (type == ID_SEND && tag == SEND_BYTES) {
                ret = Comm.writeBytesSocketConn(res2);
            } else {
	        ret = Comm.writeSocketConn(res2);
            }
            if (ret != EXIT_OK) {
	        theLog->debug("Fails in write(Bytes)SocketConn()");
            }
            // Store last results
            if (timer != NULL) {
                timer->lastValue = res2;
            }
        }
    }
       
    return EXIT_OK;
}

int AR_Client::execCmd(int cmdType, QString commandIn)
{
    int ret;
    QString commandOrig;

    if (commandIn.isNull() || commandIn.isEmpty()) {
        theLog->info("execCmd >no params<");
        commandOrig = QString::null;
    } else {
        theLog->info("execCmd >" + commandIn + "<");
        commandOrig = commandIn;
    }    
    
    switch (cmdType) {
    
        case ID_EXIT:               
    	    if (remoteOn == 1) {	    // User asks to exit	
    	        theLog->info("Exit key detected");
    	        Disconnect();			    

    	        return EXIT_ABORT;
            }
            break;
            
        case ID_TMPDISCONN:
            if (remoteOn == 1) {            // User asks to temporary disconnect
    	        theLog->info("Temporary disconnect key detected");
	        Disconnect();
                
    	        return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
            }
            break;
            
        case ID_REMOTEON:
            theLog->info("Switch remote on");
    	    remoteOn = 1;
            break;
            
        case ID_REMOTEOFF:
            theLog->info("Switch remote off");
    	    remoteOn = 0;
            break;
            
        case ID_EXEC:
            theLog->info("Command: Exec");
            if (remoteOn == 1) {
                execSimpleCmd(commandOrig, 0);
            }
            break;
            
        case ID_SETMODE:
            theLog->info("Command: SetMode");
            if (remoteOn == 1) {           
                conf->setMode(commandOrig);
            }
            break;

        case ID_SENDCKPD:
            theLog->info("Command: SendCKPD");
            if (serverMode == SERVER_NO && remoteOn == 1) {

                ret = Comm.sendSeq(commandOrig);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in sendSeq()");
                }
    
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	            return ret;
                }
            }    
            break;
            
        case ID_SEND:
            theLog->info("Command: Send");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != QString::null) {
                ret = processSendCmd(commandOrig);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in processSendCmd()");
                }
            }    
            break;
            
        case ID_SET:          
            theLog->info("Command: Set");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != QString::null) {
	    	QString tmptext(CMD_SET);
                tmptext += QString("(") + commandOrig + ");";
               
                 ret = Comm.writeSocketConn(tmptext);
                if (ret != EXIT_OK) {
	            theLog->debug("Fails in writeSocketConn() (Set)");
                }
            }    
            break;
            
        case ID_EXECSET:   
            theLog->info("Command: ExecAndSet");
            
            execCmdAndGetResults(ID_SET, commandOrig, NULL);
            break;
            
        case ID_EXECSEND:   
            theLog->info("Command: ExecAndSend");
            
            execCmdAndGetResults(ID_SEND, commandOrig,NULL);
            break;

        case ID_TIMERCMD:   
            theLog->info("Command: Timer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != QString::null) {
                return processTimerCmd(commandOrig);
            }     
            break;

        case ID_CANCELTIMER:   
            theLog->info("Command: CancelTimer");
            if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != QString::null) {
                cancelTimer(commandOrig);
            }  
            break;
            
        default:  
            theLog->info("Unknown command");                 
            break;
   }
   
   return EXIT_OK;
}

bool AR_Client::stillPressed(void)
{
    return repeatNow;
}

void AR_Client::setRemoteOn(bool set)
{
    theLog->info(QString("Switch remote ") + (set ? "On" : "Off"));
    remoteOn = set;
}

int AR_Client::checkCallActive()
{
    theLog->debug("AR_Client::checkCallActive()");
       
    // Is call still active (timeout about 1 seconds) ?
    if (callOn) {
    	theLog->debug("Get call list again");
    	
    	int ret = Comm.getClip(callerId);	    // Will set global callerId
    	if (ret == EXIT_EXACT) {
    	    theLog->debug("Call still active (" + callerId + ")");
    	    return EXIT_OK;
     	} else {
    	    theLog->debug("Call was finished");
	    callOn  	= 0;
    	    callerId	= QString("NO CALLER ID");
			    
    	    type_key* tk = conf->findItem(EVT_ENDCALL, ret);
    	    if (tk != NULL) {
    	    	theLog->debug("EVT_ENDCALL found");
    	    	ret = handleCmdByKey(tk);
    	    	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
    	    	    return ret;
    	    	}
    	    }
   	}
    }
    return EXIT_NOK;
}

//
// Functions related to timer commands
//
int AR_Client::processTimerCmd(QString origCmd)
{
    int ret = EXIT_OK;
    
    theLog->debug("processTimerCmd(): >" + origCmd + "<");
    
    QString all = origCmd;
    
    QString id  = origCmd.section(",",0,0).simplifyWhiteSpace();
    QString t1  = origCmd.section(",",1,1).simplifyWhiteSpace();
    QString t2  = origCmd.section(",",2,2).simplifyWhiteSpace();
    QString cmd = origCmd.section(",",3).simplifyWhiteSpace();

    if (id.isNull() || cmd.isNull() || t1.isNull() || t2.isNull()) {
        theLog->debug("processTimerCmd(): incorrect input");                
        return EXIT_NOK;
    }
    // Do not allow timer commands for Send(string/bytes...)
    if (!(cmd.startsWith(SET_ICONS) ||
          cmd.startsWith(SET_TITLE) ||
          cmd.startsWith(SET_STATUS))) {
        theLog->debug("processTimerCmd(): incorrect tag in command");                
        return EXIT_NOK;
    }

    int   tmout = t1.toInt();
    int   times = t2.toInt();

    if (createTimer(id, ID_SET, tmout, times, cmd) == EXIT_NOK) {
        theLog->debug("processTimerCmd(): timer already exists or input was incorrect");                
        return EXIT_NOK;
    }
    
    timerCmd* tt = NULL;
    execCmdAndGetResults(ID_SET, cmd, findTimer(id, &tt));
	
    // AR_Client is not successor of QObject so send timeout to mainWin
    theLog->debug("Set commands timer");
    QTimer::singleShot(1000, mainWin, SLOT(handleTimerCmd()));

    return ret;
}

timerCmd* AR_Client::findTimer(QString id, timerCmd **prev) 
{
    timerCmd* ptr = timers;
    
    while (ptr != NULL && (!ptr->id.startsWith(id))) {
        *prev = ptr;
        theLog->debug("findTimer(): set prev to >" + (*prev)->id + "<");
        ptr  = ptr->next;
    }

    return ptr;
}

int AR_Client::createTimer(QString id, int type, int tmout, int times, QString cmd)
{
    if (id == NULL || cmd == NULL) {
        theLog->debug("createTimer(): null input");                
        return EXIT_NOK;
    }
    
    timerCmd* ptr = NULL;
    if (findTimer(id, (timerCmd**) &ptr) != NULL) {
        theLog->debug("createTimer(): timer already exists");                
        return EXIT_NOK;
    } 
    timerCmd *tm = new timerCmd();
    
    tm->id        = id;
    tm->type      = type;
    tm->ticks     = 0;
    tm->times     = 1;
    tm->tmout     = tmout;
    tm->maxTimes  = times;
    tm->cmd       = cmd;
    tm->lastValue = QString::null;
    
    // Insert in head
    if (timers == NULL) { // Insert first
        tm->next = NULL;
    } else {
        tm->next = timers;
    }
    timers = tm;
    
    return EXIT_OK;
}

int AR_Client::cancelTimer(QString id) 
{
    theLog->debug("cancelTimer() >" + id + "<");
    
    timerCmd *prev = NULL;
    timerCmd *tm = findTimer(id, (timerCmd **) &prev);
    if (tm == NULL) {
        theLog->debug("Cant find timer to cancel!");
        return EXIT_NOK;
    }    
    if (tm == timers) { //remove in head
        timers = tm->next;
    } else {
        if (prev != NULL) {
            prev->next = tm->next;
        } else {
            theLog->error("Previous item absent for non-first element");
            return EXIT_NOK;
        }
    }        
    delete tm;
    
    return EXIT_OK;
}

void AR_Client::freeTimers()
{
    theLog->debug("freeTimers()");
    timerCmd *tm;
    
    while (timers != NULL) {
        tm     = timers;
        timers = timers->next;
        delete tm;
    }
    return;
}

int AR_Client::verifyTimers(int ticksInSec) 
{
    timerCmd *tm   = timers;
    
    int numTimers = 0;
    while (tm != NULL) {
        
        if (tm->ticks >= (tm->tmout * ticksInSec)) {
            theLog->debug("verifyTimers(): it is time to execute >" + tm->cmd + "<");

            execCmdAndGetResults(tm->type, tm->cmd, tm);
            
            tm->ticks = 0;
            
            if (tm->maxTimes > 0) {
                tm->times++;
                if (tm->times >= tm->maxTimes) {    // Cancel this timer 
                    theLog->debug("verifyTimers(): timer " + tm->id + " self-canceled");
                    
                    // Cancel this timer
                    timerCmd *tmc = tm;
                    
                    tm = tm->next;

                    cancelTimer(tmc->id);
                    
                    continue;
                }
            }
        } else {
            tm->ticks++;
        }
        //prev = tm;
        tm   = tm->next;
        numTimers++;
    }
    return numTimers;
}
//
// Bemused server limited emulation
//
bool  AR_Client::isBemusedCommand(char *cmdIn)
{  
    theLog->debug("isBemusedCommand()");
    
    bool isBemused = false;
    
    char *cmd = cmdIn;
    char *last = cmdIn + strlen(cmdIn);
    
    while (cmd < last) {
    	int handle = 0;
	int shift  = 4;
	char oneCmd[5];
	
    	if (strncmp(cmd, "CHCK", 4) == 0 ||
	    strncmp(cmd, "DINF", 4) == 0 ||
    	    strncmp(cmd, "EXIT", 4) == 0 ||
    	    strncmp(cmd, "FADE", 4) == 0 ||
    	    strncmp(cmd, "FFWD", 4) == 0 ||
	    strncmp(cmd, "GVOL", 4) == 0 ||
	    strncmp(cmd, "INF2", 4) == 0 ||
	    strncmp(cmd, "INFO", 4) == 0 ||
	    strncmp(cmd, "LIST", 4) == 0 ||
    	    strncmp(cmd, "NEXT", 4) == 0 ||
    	    strncmp(cmd, "PAUS", 4) == 0 ||
	    strncmp(cmd, "PLEN", 4) == 0 ||
	    strncmp(cmd, "PLST", 4) == 0 ||
    	    strncmp(cmd, "PREV", 4) == 0 ||
	    strncmp(cmd, "RMAL", 4) == 0 ||
    	    strncmp(cmd, "RWND", 4) == 0 ||
	    strncmp(cmd, "SHUT", 4) == 0 ||
    	    strncmp(cmd, "STEN", 4) == 0 ||
    	    strncmp(cmd, "STOP", 4) == 0 ||
            strncmp(cmd, "STRT", 4) == 0 ||
	    strncmp(cmd, "VERS", 4) == 0) {
	    handle = 1;
    	}

    	if (strncmp(cmd, "REPT", 4) == 0 ||
            strncmp(cmd, "SHFL", 4) == 0) {
            unsigned char bc = *((unsigned char*) (cmd + 4));
            paramValue = (int)bc;

            handle = 1;
	    shift  = 5;
    	}
    	if (strncmp(cmd, "VOLM", 4) == 0) {
	    unsigned char bc;
	    if (strlen(cmd) == 4) {  // read only VOLM without value to set
	    	theLog->debug("read one more byte");
		char ret[2];
		ret[1] = '\0';;
	    	if ((ret[0] = (char) Comm.readchar(500000)) < 0) {
            	    theLog->debug("isBemusedCommand: read < 0");
		    paramValue = -1;
		    return 0;
		}
		QString ch(ret);
            	theLog->debug("isBemusedCommand: read >" + QString(ret) + "<");
		bc = (unsigned char) ret[0];
	    } else {
	    	bc = ((unsigned char) *(cmd + 4));
	    }
 
    	    // VOLM 0-255 -> 0%-100%
            paramValue = ((int)bc * 100)/255;
        
    	    theLog->debug("isBemusedCommand: parameter >" + QString().setNum(paramValue) + "<");	    
	    
	    handle = 1;
	    shift  = 5;
    	}
    	if (strncmp(cmd, "SLCT", 4) == 0) {
            handle = 1;
	    shift  = 6;
    	}
    	if (strncmp(cmd, "SEEK", 4) == 0)  {
	    paramValue = (((unsigned char)*(cmd + 4)) << 24) + 
	                 (((unsigned char)*(cmd + 5)) << 16) + 
		         (((unsigned char)*(cmd + 6)) << 8)  + 
		           (unsigned char)*(cmd + 7);
            handle = 1;
	    shift  = 8;
    	}
    	if (strncmp(cmd, "DLST", 4) == 0 || 
            strncmp(cmd, "DOWN", 4) == 0 || 
            strncmp(cmd, "FINF", 4) == 0 ||    
    	    strncmp(cmd, "LADD", 4) == 0 ||
    	    strncmp(cmd, "PLAY", 4) == 0) {             
	    
            unsigned char bc = ((unsigned char) *(cmd + 4));
	    
            handle = 1;
	    shift  = 5 + bc;
    	}
    	if (handle) {
	    isBemused = true;
	    strncpy(oneCmd,cmd,4);
	    oneCmd[4] = '\0';
           
   	    theLog->debug("isBemusedCommand: one commmand >)" + QString(oneCmd) + "<");	    
    	    
    	    handleKeyPress(oneCmd);
    	} else {
	    break;
	}
	
	cmd = cmd + shift;
    }
    return isBemused;
}
